from .base import *
from .gemini import *
from .syosetu_com import *
from .syosetu_org import *
from .deepl import *
from .gpt import *
from .autocrawler import *
from .epub import *
from .esj import *
from .masiro import *
from .kakuyomu import *
from .novelup import *


from __future__ import annotations

import inspect
from importlib import import_module
from abc import abstractmethod, ABC
from typing import AsyncGenerator, Any, TypeVar
from uuid import uuid4
from pydantic import Field, BaseModel, ConfigDict, computed_field
from ..classes import Context, Book, Chapter, Episode, Task


R = TypeVar('R', bound='BaseRole')


__all__ = [
    "BaseRole",
    "BaseCrawler",
    "BaseTranslator",
    "BaseExporter",
    "recover_from_dict",
]


class BaseRole(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str = Field(default_factory=lambda: f"role-{str(uuid4())[:8]}")
    description: str = ''

    @computed_field
    def classname(self) -> str:
        return self.__class__.__name__

    @computed_field
    def modulename(self) -> str | None:
        mod = inspect.getmodule(self.__class__)
        if mod is None:
            return None
        return mod.__name__

    def __str__(self) -> str:
        return f"{self.classname}({self.name})"

    def __repr__(self) -> str:
        return f"{self.classname}({self.name!r})"


def recover_from_dict(
        d: dict[str, Any],
        type_: type[R] = BaseRole) -> R:  # type: ignore[assignment]
    """
    Recover a BaseRole object from a dictionary
    """
    try:
        mod = import_module(d['modulename'])
        cls = getattr(mod, d['classname'])
        if not issubclass(cls, type_):
            raise ValueError(f"Class {cls} is not a subclass of {type_}")
        return cls(**d)
    except (KeyError, AttributeError, ImportError) as e:
        raise ValueError(f"Failed to recover from dict: {e}") from e


class BaseCrawler(ABC, BaseRole):
    """
    Crawler for syosetu.com
    """
    name: str = Field(default_factory=lambda: f"crawler-{str(uuid4())[:8]}")
    description: str = "Base Crawler"

    @abstractmethod
    def crawl_stream(
            self,
            task: Task,
            ctx: Context,
    ) -> AsyncGenerator[tuple[Task, Book, Chapter, Episode], None]:
        ...


class BaseTranslator(ABC, BaseRole):
    """
    Translator base class
    """
    name: str = Field(default_factory=lambda: f"translator-{str(uuid4())[:8]}")
    description: str = "Base Translator"

    @abstractmethod
    async def translate(
            self,
            episode: Episode,
            task: Task,
            ctx: Context,
            chapter: Chapter | None = None,
            book: Book | None = None,
    ) -> Episode:
        ...


class BaseExporter(ABC, BaseRole):
    """
    Exporter base class
    """
    name: str = Field(default_factory=lambda: f"exporter-{str(uuid4())[:8]}")
    description: str = "Base Exporter"

    @abstractmethod
    async def export(
            self,
            book: Book,
            task: Task,
            ctx: Context,
    ) -> Any:
        ...



from __future__ import annotations

import httpx
import asyncio
from typing import Any
from pydantic import BaseModel, field_validator, ConfigDict
from pydantic import Field, PrivateAttr
from asynctinydb import TinyDB
from datetime import datetime, UTC
from itertools import chain
from vermils.gadgets import LoggerLike


__all__ = [
    "TimeMeta",
    "ImageRef",
    "Line",
    "Episode",
    "Chapter",
    "Book",
    "Context",
    "Task"
]


class TimeMeta(BaseModel):
    """
    # TimeMeta Class
    """
    model_config = ConfigDict(validate_assignment=True)
    created_at: datetime | None = None
    updated_at: datetime | None = None
    saved_at: datetime = Field(default_factory=datetime.utcnow)

    @field_validator("created_at", "updated_at", "saved_at", mode="after")
    def convert_utc(cls, v: datetime | None):
        if v is not None:
            return v.astimezone(UTC)
        return v


class ImageRef(BaseModel):
    """
    # ImageRef Class
    """
    url: str
    alt: str = ''
    time_meta: TimeMeta = Field(default_factory=TimeMeta)


class Line(BaseModel):
    """
    # Line Class
    """
    content: str
    translated: str | None = None
    candidates: dict[str, str] = Field(default_factory=dict)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Line):
            return self.content == other.content
        return NotImplemented

    def __str__(self) -> str:
        return self.content

    def __repr__(self) -> str:
        return f"Line(content={self.content!r}, translated={self.translated!r})"


class Episode(BaseModel):
    """
    # Episode Class
    """
    title: str
    title_translated: str | None = None
    notes: str = ''
    notes_translated: str | None = None
    lines: list[Line]
    time_meta: TimeMeta = Field(default_factory=TimeMeta)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Episode):
            return self.title == other.title
        return NotImplemented

    @property
    def fully_translated(self) -> bool:
        """
        # Fully Translated
        Check if all lines are translated
        """
        for line in self.lines:
            if not line.content.strip():
                continue
            if line.translated is None:
                return False
        return True

    @property
    def html(self) -> str:
        """
        # HTML
        """
        return '\n'.join(chain(
            (f"<h3>{self.title_translated or self.title}</h3>",
             f"<p>{self.notes_translated or self.notes}</p>",
             "<hr>"),
            (f"<p>{line.translated or ''}</p>" for line in self.lines)
        ))

    @property
    def raw_html(self) -> str:
        """HTML of the original content"""
        return '\n'.join(chain(
            (f"<h3>{self.title}</h3>",
             f"<p>{self.notes}</p>",
             "<hr>"),
            (f"<p>{line.content}</p>" for line in self.lines)
        ))


class Chapter(BaseModel):
    """
    # Chapter Class
    """
    title: str
    title_translated: str | None = None
    cover: ImageRef | None = None
    episodes: list[Episode] = Field(default_factory=list)
    time_meta: TimeMeta = Field(default_factory=TimeMeta)

    def get_episode(self, title: str) -> Episode | None:
        """
        # Get Episode
        """
        for episode in self.episodes:
            if episode.title == title:
                return episode
        return None

    @property
    def fully_translated(self) -> bool:
        return all(episode.fully_translated for episode in self.episodes)

    @property
    def html(self) -> str:
        return '\n'.join(chain(
            (f"<h2>{self.title_translated or self.title}</h2>",),
            (episode.html for episode in self.episodes)
        )
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Chapter):
            return self.title == other.title
        return NotImplemented


class Book(BaseModel):
    """
    # Book Class
    """
    title: str
    title_translated: str | None = None
    author: str
    url: str | None = None
    series: str | None = None
    series_translated: str | None = None
    tags: set[str] = Field(default_factory=set)
    cover: ImageRef | None = None
    description: str = ''
    description_translated: str | None = None
    time_meta: TimeMeta = Field(default_factory=TimeMeta)
    chapters: list[Chapter] = Field(default_factory=list)

    def get_chapter(self, title: str) -> Chapter | None:
        """
        # Get Chapter
        """
        for chapter in self.chapters:
            if chapter.title == title:
                return chapter
        return None

    @property
    def fully_translated(self) -> bool:
        return all(chapter.fully_translated for chapter in self.chapters)

    @property
    def html(self) -> str:
        return '\n'.join(chain(
            (f"<h1>{self.title_translated or self.title}</h1>",
             f"<p>{self.description_translated or self.description}</p>",
             "<hr>"),
            (chapter.html for chapter in self.chapters)
        ))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Book):
            return (self.title == other.title
                    and self.author == other.author)
        return NotImplemented


class Context:
    def __init__(self,
                 db: TinyDB,
                 client: httpx.AsyncClient,
                 logger: LoggerLike,
                 extra: dict[str, Any] | None = None):
        self.db = db
        self.logger = logger
        self.client = client
        self.extra = extra or {}


class Task(BaseModel):
    """
    # Task Class
    """
    url: str = ''
    friendly_name: str = ''
    sauce_lang: str = "JA"
    target_lang: str = "ZH"
    crawler: str | None = None
    translator: str | list[str] | None = None
    exporter: str | list[str] | None = None
    glossaries: list[tuple[str, str]] = Field(default_factory=list)
    extra: dict[str, Any] = Field(default_factory=dict)
    _lock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)

    @property
    def lock(self) -> asyncio.Lock:
        return self._lock



import re
import bs4
from base64 import b64decode
from vermils.io import aio
from .classes import Context


async def get_url_content(url: str, origin: str, ctx: Context) -> bytes:
    if url.startswith("http"):
        r = await ctx.client.get(
            url,
            headers={"Origin": origin},
            follow_redirects=True,
        )
        r.raise_for_status()
        data = r.content
    elif url.startswith("base64://"):
        data = b64decode(url.removeprefix("base64://"))
    elif url.startswith("file://"):
        async with aio.open(url.removeprefix("file://"), "rb") as f:
            data = await f.read()
    else:
        raise ValueError(f"Cannot download URL: {url}")
    return data

_ruby_re = re.compile(r"<ruby>.*?<\/ruby>")
_un_ruby_re = re.compile(r"\s?\[.*?\]\(\^.*?\)\s?")


def escape_ruby(s: str) -> str:
    ruby_matches = _ruby_re.findall(s)
    for match in ruby_matches:
        base = ''
        top = ''
        soup = bs4.BeautifulSoup(match, "xml")
        for child in soup.ruby.children:
            if isinstance(child, bs4.NavigableString):
                base += child.strip('\n')
            elif child.name == "rb":
                base += child.text.strip('\n')
            elif child.name == "rt":
                top += child.text.strip('\n')
        replacement = f" [{base}](^{top}) "
        s = s.replace(match, replacement)
    return s


def unescape_ruby(s: str) -> str:
    ruby_matches = _un_ruby_re.findall(s)
    for match in ruby_matches:
        base: str
        top: str
        base, top = match.strip(' ')[1:-1].split("](^")
        base, top = base.strip(' '), top.strip(' ')
        replacement = f"<ruby><rb>{base}</rb><rt>{top}</rt></ruby>"
        s = s.replace(match, replacement)
    return s



from __future__ import annotations

from uuid import uuid4
from pydantic import Field
from asynctinydb import Query
from aiodeepl import Translator
from .base import BaseTranslator
from ..classes import Book, Chapter, Episode, Task, Context


__all__ = ["DeepLTranslator"]


class DeepLTranslator(BaseTranslator):
    """
    # DeepLTranslator Class
    """
    name: str = Field(default_factory=lambda: f"deepl-{str(uuid4())[:8]}")
    description: str = "DeepL Translator"
    skip_translated: bool = True
    backend: Translator = Field(
        default_factory=lambda: Translator(api_key=''))

    async def translate(
            self,
            episode: Episode,
            task: Task,
            ctx: Context,
            chapter: Chapter | None = None,
            book: Book | None = None,
    ) -> Episode:
        logger = ctx.logger
        logger.info("%s: Translating episode %s", self, episode.title)
        indexes = list[int]()
        contents = list[str]()
        glossaries = task.glossaries
        sauce_lang = task.sauce_lang
        target_lang = task.target_lang
        for i, line in enumerate(episode.lines):
            if not line.content.strip():
                continue  # skip empty lines
            if self.skip_translated and line.translated is not None:
                continue
            indexes.append(i)
            contents.append(line.content)
        if not indexes:
            logger.info("%s: Episode %s is fully translated", self, episode.title)
            return episode
        logger.info("%s translating %d lines", self, len(indexes))
        context = ''
        if book is not None:
            context += f"Book title: {book.title}\n"
            if book.description:
                context += f"Book description: {book.description}\n"
            if book.tags:
                context += f"Book tags: {','.join(book.tags)}\n"
        if chapter is not None:
            context += f"Chapter title: {chapter.title}\n"
        if episode.title:
            context += f"Episode title: {episode.title}\n"
        gid: str | None = None
        if glossaries:
            pairs = await self.backend.glossary_available_pairs()
            if (sauce_lang, target_lang) in pairs:
                logger.debug("%s: Glossary pair uploading", self)
                gid = (await self.backend.glossary_create(
                    f"glossary-{str(uuid4())[:8]}",
                    sauce_lang,
                    target_lang,
                    entries=glossaries,
                )).glossary_id
            else:
                logger.warning("%s: Glossary pair %s-%s not available",
                               self, sauce_lang, target_lang)

        try:
            async def translate(c: list[str] | str):
                return await self.backend.translate(
                    c,
                    target_lang=target_lang,
                    source_lang=sauce_lang,
                    context=context,
                    split_sentences='0',
                    preserve_formatting=True,
                    tagged_handling="html",
                    glossary_id=gid,
                )

            if book is not None:
                if book.title and (book.title_translated is None or not self.skip_translated):
                    book.title_translated = (await translate(book.title)).text
                if book.description and (book.description_translated is None or not self.skip_translated):
                    book.description_translated = (await translate(book.description)).text
                if book.series and (book.series_translated is None or not self.skip_translated):
                    book.series_translated = (await translate(book.series)).text

            if ((chapter is not None and chapter.title)
                    and (chapter.title_translated is None or not self.skip_translated)):
                chapter.title_translated = (await translate(chapter.title)).text

            if episode.title and (episode.title_translated is None or not self.skip_translated):
                episode.title_translated = (await translate(episode.title)).text
            if episode.notes and (episode.notes_translated is None or not self.skip_translated):
                episode.notes_translated = (await translate(episode.notes)).text

            translated = await translate(contents)

            for i, t in zip(indexes, translated):
                logger.debug("%s %s -> %s", self, episode.lines[i].content, t.text)
                episode.lines[i].translated = t.text
                if not t.text.strip():
                    logger.warning("%s: Empty translation for line %s in episode %s",
                                   self, episode.lines[i].content, episode.title)

            logger.info("%s: Translated episode %s", self, episode.title)

            if book is not None:
                logger.debug("%s: Upserting book %s", self, book.title)
                q = Query()
                await ctx.db.upsert(book.model_dump(mode="json"),
                                    q.title == book.title and q.author == book.author)
            return episode
        finally:
            if gid:
                await self.backend.glossary_delete(gid)
